//
// Copyright 2006 (C) Eiichiro ITO, GHC02331@nifty.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eiichiro ITO, 1 Jun 2006
// mailto: GHC02331@nifty.com
//
// 2006.6.1 Originally Created.
//

#include <string.h>
#include <OPENR/OSyslog.h>
#include <OPENR/OPENRAPI.h>
#include <ant.h>
#include <EndpointTypes.h>
#include <TCPEndpointMsg.h>
#include <OPENR/core_macro.h>
#include "TCPReceptor.h"

#define AIBO_MAGIC "AIBO"
#ifdef ERS210
#define AIBO_TAG "21"
#endif /* ERS210 */
#ifdef ERS7
#define AIBO_TAG "07"
#endif /* ERS7 */
#define HEADER_LENGTH (4+2+8)

void
TCPReceptor::SendOkWithTagString( int index, const char *tag, const char *string )
{
    int len;

    len = strlen(string) + 2;
    sprintf( (char*) connection[index].sendData, "%s%s%08d%-2.2s%s",
	     AIBO_MAGIC, AIBO_TAG,
	     len, tag, string );
    connection[index].sendSize = HEADER_LENGTH + len;
    Send( index );
}

void
TCPReceptor::SendOkWithTagData( int index, const char *tag, const byte *data, int size )
{
    sprintf( (char*) connection[index].sendData,
	     "%s%s%08d%-2.2s",
	     AIBO_MAGIC, AIBO_TAG,
	     size + 2, tag );
    memcpy( &connection[index].sendData[HEADER_LENGTH + 2], data, size );

    connection[index].sendSize = HEADER_LENGTH + 2 + size;
    Send( index );
}

int
from2digits( const char *buf )
{
    char number[ 3 ];
    memcpy( number, buf, 2 );
    number[ 2 ] = 0;
    return atoi( number );
}

int
from4digits( const char *buf )
{
    char number[ 5 ];
    memcpy( number, buf, 4 );
    number[ 4 ] = 0;
    return atoi( number );
}

void
TCPReceptor::ProcessReceivedMessage( int index, const byte* buf, int size )
{
    ControlInfo ci;
    char tmpbuf[32];

    OSYSDEBUG(("TCPReceptor::ProcessReceivedMessage code=[%-2.2s]\n", buf));

    memset( &ci, 0, sizeof ci );
    if ( !memcmp( buf, "MO", 2 ) ) {
	ci.id = CtID_Monet;
	ci.iValue = from4digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "MS", 2 ) ) {
	ci.id = CtID_MonetSync;
	ci.iValue = from4digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "ST", 2 ) ) {
	ci.id = CtID_StartSTN;
	ci.iValue = from2digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SB", 2 ) ) {
	ci.id = CtID_StopSTN;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SR", 2 ) ) {
	ci.id = CtID_ReadSTN;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SS", 2 ) ) {
	ci.id = CtID_GetCurrentStateSTN;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	connectionIndex = index;
    } else if ( !memcmp( buf, "GC", 2 ) ) {
	imageType = IMAGE_LAYER_C;
	connectionIndex = index;
        observer[obsImage]->AssertReady();
    } else if ( !memcmp( buf, "GL", 2 ) ) {
	imageType = IMAGE_LAYER_L;
	connectionIndex = index;
        observer[obsImage]->AssertReady();
    } else if ( !memcmp( buf, "GM", 2 ) ) {
	imageType = IMAGE_LAYER_M;
	connectionIndex = index;
        observer[obsImage]->AssertReady();
    } else if ( !memcmp( buf, "GI", 2 ) ) {
	ci.id = CtID_GetCdtInfo;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	connectionIndex = index;
    } else if ( !memcmp( buf, "CG", 2 ) ) {
	ci.id = CtID_GetCdt;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	connectionIndex = index;
    } else if ( !memcmp( buf, "CR", 2 ) ) {
	ci.id = CtID_ReadCdt;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "HD", 2 ) ) {
	// 0123456789012345678901234
	// HDXXX,XXX,-XX,-XX,-XX,-XX
	ci.id = CtID_Head;
	char arg[ 4 ];
	memcpy( arg, &buf[2], 3 );  arg[3] = '\0'; // XXX(Command)
	ci.iValue = atoi(arg);
	memcpy( arg, &buf[6], 3 );  arg[3] = '\0'; // XXX(Arg1)
	ci.i2Value = atoi(arg);
	memcpy( arg, &buf[10], 3 );  arg[3] = '\0'; // -XX(TILT)
	ci.dValue2[0] = atof(arg);
	memcpy( arg, &buf[14], 3 ); arg[3] = '\0'; // -XX(PAN)
	ci.dValue2[1] = atof(arg);
	memcpy( arg, &buf[18], 3 ); arg[3] = '\0'; // -XX(TILT2)
	ci.dValue2[2] = atof(arg);
	memcpy( arg, &buf[22], 3 ); arg[3] = '\0'; // -XX(MOUTH)
	ci.dValue2[3] = atof(arg);
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "HR", 2 ) ) {
	ci.id = CtID_ReadHead;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "FC", 2 ) ) {
	ci.id = CtID_Face;
	ci.iValue = from2digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SP", 2 ) ) {
	// SP-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,\\
	// -XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,
	ci.id = CtID_SetPose;
	const byte *ptr = &buf[2];
	char arg[10];
	for ( int i = 0; i < 12; i += 1 ) {
	    memcpy( arg, ptr, 7 );  // -XXX.XX
	    arg[7] = 0;
	    ci.dValue2[i] = atof(arg);
	    ptr += 8;
	}
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SC", 2 ) ) {
	ci.id = CtID_ClearSTN;
	ci.iValue = from2digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SA", 2 ) ) {
	ci.id = CtID_AddSTN;
	ci.iValue = from2digits( (char*) &buf[2] );
	strncpy( ci.sValue, (char*) &buf[4], sizeof ci.sValue );
	char *p = strchr( ci.sValue, '$' );
	if ( p ) {
	    *p = 0;
	}
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "CS", 2 ) ) {
	ci.id = CtID_SendCdt;
	strncpy( ci.sValue, (char*) &buf[2], sizeof ci.sValue );
	char *p = strchr( ci.sValue, '$' );
	if ( p ) {
	    *p = 0;
	}
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "CW", 2 ) ) {
	ci.id = CtID_WriteCdt;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "IN", 2 ) ) {
	ci.id = CtID_Internal;
	ci.iValue = from2digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "$$", 2 ) ) {
	connection[index].receiveNext = false;
	SendOkWithTagString( index, "OK", "" );
    } else {
	OSYSPRINT(( "%s unknown command: (%-10.10s)\n",
		    "TCPReceptor::ProcessReceivedMessage",
		    buf ));

	SendOkWithTagString( index, "NG", "" );
    }
}

void
TCPReceptor::NotifyImage(const ONotifyEvent& event)
{
    // Visionǡä
    if ( tcpReceptorState == TCPRS_IDLE ) {
        return;
    }
    OSYSDEBUG(("TCPReceptor::NotifyImage(%d)\n", connectionIndex));
    if ( connectionIndex < 0 ) {
	return;
    }
    connection[connectionIndex].receiveNext = false;

    byte* data;
    switch ( imageType ) {
    case IMAGE_LAYER_C:
	data = (byte*) event.Data(0);
	SendOkWithTagData( connectionIndex, "IC", data, LAYER_C_IMAGESIZE );
	break;

    case IMAGE_LAYER_L:
	data = (byte*) event.Data(1);
	SendOkWithTagData( connectionIndex, "IL", data, LAYER_L_IMAGESIZE );
	break;

    case IMAGE_LAYER_M:
	data = (byte*) event.Data(2);
	SendOkWithTagData( connectionIndex, "IM", data, LAYER_M_IMAGESIZE );
	break;
    }
    connectionIndex = -1;
}

void
TCPReceptor::NotifyControlResult( const ONotifyEvent& event )
{
    // Brainä
    if ( tcpReceptorState == TCPRS_IDLE ) {
        return;
    }
    OSYSDEBUG(("TCPReceptor::NotifyControlResult(%d)\n", connectionIndex));
    if ( connectionIndex < 0 ) {
	return;
    }
    connection[connectionIndex].receiveNext = false;

    ControlInfo *cr = (ControlInfo*) event.Data(0);
    switch ( cr->id ) {
    case CtID_GetCurrentStateSTN:
	SendOkWithTagString( connectionIndex, "SS", cr->sValue );
	break;
    case CtID_GetCdtInfo:
	SendOkWithTagString( connectionIndex, "BI", cr->sValue );
	break;
    case CtID_GetCdt:
	SendOkWithTagString( connectionIndex, "CI", cr->sValue );
	break;
    }
    observer[event.ObsIndex()]->AssertReady();
    connectionIndex = -1;
}
